package org.openmuc.jdlms.internal;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.openmuc.jdlms.AccessResultCode;
import org.openmuc.jdlms.AuthenticationMechanism;
import org.openmuc.jdlms.CosemAttribute;
import org.openmuc.jdlms.CosemClass;
import org.openmuc.jdlms.CosemInterfaceObject;
import org.openmuc.jdlms.IllegalAttributeAccessException;
import org.openmuc.jdlms.IllegalMethodAccessException;
import org.openmuc.jdlms.LogicalDevice;
import org.openmuc.jdlms.MethodAccessMode;
import org.openmuc.jdlms.MethodResultCode;
import org.openmuc.jdlms.ObisCode;
import org.openmuc.jdlms.SecuritySuite;
import org.openmuc.jdlms.SecuritySuite.SecurityPolicy;
import org.openmuc.jdlms.SelectiveAccessDescription;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.datatypes.DataObject.Type;

public class DataDirectory {

    private final Map<Integer, DlmsLogicalDevice> logicalDeviceMap;

    private final Map<Long, ServerConnectionData> connectionsData;

    public DataDirectory() {
        this.logicalDeviceMap = new HashMap<>();
        this.connectionsData = new HashMap<>();
    }

    public synchronized DataObject invokeMethod(int logicalDeviceId, ObisCode instanceId, long classId, long methodId,
            DataObject param, Long connectionId) throws IllegalMethodAccessException {
        CosemClassInstance dlmsClassInstance = retrieveDlmsClassInstance(logicalDeviceId, instanceId, (int) classId);

        if (dlmsClassInstance == null) {
            throw new IllegalMethodAccessException(MethodResultCode.OBJECT_UNDEFINED);
        }

        MethodAccessor method = findMethod(methodId, dlmsClassInstance);
        SecuritySuite sec = getConnectionData(connectionId).securitySuite;
        SecurityPolicy securityPolicy = sec.getSecurityPolicy();

        MethodAccessMode accessMode = method.getDlmsMethod().accessMode();

        if (accessMode == MethodAccessMode.NO_ACCESS) {
            throw new IllegalMethodAccessException(MethodResultCode.READ_WRITE_DENIED);
        }

        if (accessMode == MethodAccessMode.AUTHENTICATED_ACCESS && !securityPolicy.isAuthenticated()) {
            throw new IllegalMethodAccessException(MethodResultCode.READ_WRITE_DENIED);
        }

        return method.invoke(dlmsClassInstance, param, connectionId, securityPolicy);
    }

    private MethodAccessor findMethod(long methodId, CosemClassInstance dlmsClassInstance)
            throws IllegalMethodAccessException {
        if (dlmsClassInstance == null) {
            throw new IllegalMethodAccessException(MethodResultCode.OBJECT_UNDEFINED);
        }
        return dlmsClassInstance.getMethod((byte) methodId);
    }

    public Set<Integer> getLogicalDeviceIds() {
        return this.logicalDeviceMap.keySet();
    }

    public boolean doesLogicalDeviceExists(int logicalDeviceId) {
        return this.logicalDeviceMap.containsKey(logicalDeviceId);
    }

    public DlmsLogicalDevice addLogicalDevice(int logicalDeviceId, DlmsLogicalDevice logicalDevice) {
        return this.logicalDeviceMap.put(logicalDeviceId, logicalDevice);
    }

    public synchronized AccessResultCode set(int logicalDeviceId, DataObject dataObject, ObisCode instanceId,
            long classId, long attributeId, SelectiveAccessDescription selectiveAccessDescription, Long connectionId) {

        Attribute entry;
        CosemClassInstance dlmsClassInstance;
        try {
            dlmsClassInstance = retrieveDlmsClassInstance(logicalDeviceId, instanceId, (int) classId);
            if (dlmsClassInstance == null) {
                return AccessResultCode.OBJECT_UNDEFINED;
            }

            entry = retrieveAttribute(dlmsClassInstance, (int) attributeId);
        } catch (IllegalAttributeAccessException e) {
            return e.getAccessResultCode();
        }

        CosemAttribute attributeProperties = entry.attributeProperties;

        try {
            checkSetAccess(attributeProperties, dataObject.getType());
        } catch (IllegalAttributeAccessException e) {
            return e.getAccessResultCode();
        }

        ServerConnectionData connectionData = getConnectionData(connectionId);
        try {
            entry.accessor.set(dataObject, dlmsClassInstance, selectiveAccessDescription, connectionId,
                    connectionData.securitySuite.getSecurityPolicy());
        } catch (IllegalAttributeAccessException e) {
            return e.getAccessResultCode();
        }

        return AccessResultCode.SUCCESS;
    }

    private Attribute retrieveAttribute(CosemClassInstance dlmsClassInstance, int attributeId)
            throws IllegalAttributeAccessException {

        Attribute entry = dlmsClassInstance.getAttribute((byte) attributeId);

        if (entry == null) {
            throw new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
        }

        return entry;
    }

    private CosemClassInstance retrieveDlmsClassInstance(int logicalDeviceId, ObisCode instanceId, int classId) {
        DlmsLogicalDevice logicalDevice = this.logicalDeviceMap.get(logicalDeviceId);
        CosemClassInstance dlmsClassInstance = logicalDevice.get(instanceId);

        if (dlmsClassInstance == null || classId != dlmsClassInstance.getCosemClass().id()) {
            return null;
        }

        return dlmsClassInstance;
    }

    private void checkSetAccess(CosemAttribute attributeProperties, Type type) throws IllegalAttributeAccessException {
        switch (attributeProperties.accessMode()) {
        case READ_ONLY:
        case NO_ACCESS:
            throw new IllegalAttributeAccessException(AccessResultCode.READ_WRITE_DENIED);
        default:
            // TODO: Handle other cases!
            break;
        }

        if (attributeProperties.type() != Type.DONT_CARE && attributeProperties.type() != type) {
            // TODO: error wrong dataobject provided!
        }
    }

    public synchronized DataObject get(int logicalDeviceId, ObisCode instanceId, long classId, long attributeId,
            SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
            throws IllegalAttributeAccessException {
        CosemClassInstance dlmsClassInstance = retrieveDlmsClassInstance(logicalDeviceId, instanceId, (int) classId);

        ServerConnectionData connectionData = this.connectionsData.get(connectionId);
        if (dlmsClassInstance == null) {
            throw new IllegalAttributeAccessException(AccessResultCode.OBJECT_UNDEFINED);
        }

        Attribute entry = retrieveAttribute(dlmsClassInstance, (int) attributeId);

        CosemAttribute attributeProperties = entry.attributeProperties;
        checkGetAccess(attributeProperties, connectionData);

        return entry.accessor.get(dlmsClassInstance, selectiveAccessDescription, connectionId,
                connectionData.securitySuite.getSecurityPolicy());
    }

    private void checkGetAccess(CosemAttribute attributeProperties, ServerConnectionData connectionData)
            throws IllegalAttributeAccessException {
        switch (attributeProperties.accessMode()) {
        case AUTHENTICATED_WRITE_ONLY:
        case WRITE_ONLY:
        case NO_ACCESS:
            throw new IllegalAttributeAccessException(AccessResultCode.READ_WRITE_DENIED);

        case AUTHENTICATED_READ_ONLY:
        case AUTHENTICATED_READ_AND_WRITE:
            SecuritySuite securitySuite = connectionData.securitySuite;
            if (securitySuite.getAuthenticationMechanism() == AuthenticationMechanism.NONE) {
                throw new IllegalAttributeAccessException(AccessResultCode.READ_WRITE_DENIED);
            }

        case READ_AND_WRITE:
        case READ_ONLY:
        default:
            break;
        }
    }

    public static class DlmsLogicalDevice {

        private final Map<ObisCode, CosemClassInstance> classes;
        private final LogicalDevice logicalDevice;

        public DlmsLogicalDevice(LogicalDevice logicalDevice) {
            this.classes = new HashMap<>();
            this.logicalDevice = logicalDevice;
        }

        public CosemClassInstance put(ObisCode key, CosemClassInstance classInstance) {
            return this.classes.put(key, classInstance);
        }

        public LogicalDevice getLogicalDevice() {
            return logicalDevice;
        }

        public Set<ObisCode> getInstanceIds() {
            return this.classes.keySet();
        }

        public CosemClassInstance get(ObisCode key) {
            return this.classes.get(key);
        }
    }

    public static class CosemClassInstance {
        private final Map<Byte, Attribute> attributesMap;
        private final Map<Byte, MethodAccessor> methodsMap;

        private final CosemInterfaceObject instance;
        private final CosemClass cosemClass;

        public CosemClassInstance(CosemClass cosemClass, CosemInterfaceObject instance) {
            this.attributesMap = new HashMap<>();
            this.methodsMap = new HashMap<>();

            this.instance = instance;
            this.cosemClass = cosemClass;
        }

        public CosemClass getCosemClass() {
            return cosemClass;
        }

        public CosemInterfaceObject getInstance() {
            return instance;
        }

        public Attribute putAttribute(Byte attributeId, Attribute value) {
            return this.attributesMap.put(attributeId, value);
        }

        private Attribute getAttribute(Byte attributeId) {
            return this.attributesMap.get(attributeId);
        }

        public MethodAccessor putMethod(Byte methodId, MethodAccessor value) {
            return this.methodsMap.put(methodId, value);
        }

        private MethodAccessor getMethod(Byte methodId) {
            return this.methodsMap.get(methodId);
        }

        public Collection<Attribute> getAttribute() {
            return this.attributesMap.values();
        }

        public Collection<MethodAccessor> getMethods() {
            return this.methodsMap.values();
        }

    }

    public static class Attribute {
        private final AttributeAccessor accessor;
        private final CosemAttribute attributeProperties;

        public Attribute(AttributeAccessor accessor, CosemAttribute attributeProperties) {
            this.accessor = accessor;
            this.attributeProperties = attributeProperties;
        }

        public CosemAttribute getAttributeProperties() {
            return attributeProperties;
        }

    }

    public DlmsLogicalDevice getLogicalDeviceFor(Integer logicalDeviceId) {
        return this.logicalDeviceMap.get(logicalDeviceId);
    }

    public ServerConnectionData addConnection(Long connectionId, ServerConnectionData connectionData) {
        return this.connectionsData.put(connectionId, connectionData);
    }

    public ServerConnectionData getConnectionData(Long connectionId) {
        return connectionsData.get(connectionId);
    }

    public ServerConnectionData removeConnection(Long connectionId) {
        return this.connectionsData.remove(connectionId);
    }

}
