package org.openmuc.jdlms.internal.systemclasses;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import org.openmuc.jdlms.AttributeAccessMode;
import org.openmuc.jdlms.AuthenticationMechanism;
import org.openmuc.jdlms.CosemAttribute;
import org.openmuc.jdlms.CosemClass;
import org.openmuc.jdlms.CosemInterfaceObject;
import org.openmuc.jdlms.CosemMethod;
import org.openmuc.jdlms.IllegalMethodAccessException;
import org.openmuc.jdlms.LogicalDevice;
import org.openmuc.jdlms.MethodAccessMode;
import org.openmuc.jdlms.MethodResultCode;
import org.openmuc.jdlms.ObisCode;
import org.openmuc.jdlms.SecuritySuite;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.datatypes.DataObject.Type;
import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.DataDirectory.Attribute;
import org.openmuc.jdlms.internal.DataDirectory.CosemClassInstance;
import org.openmuc.jdlms.internal.DataDirectory.DlmsLogicalDevice;
import org.openmuc.jdlms.internal.MethodAccessor;
import org.openmuc.jdlms.internal.ServerConnectionData;
import org.openmuc.jdlms.internal.WellKnownInstanceIds;
import org.openmuc.jdlms.internal.security.HlsProcessorGmac;
import org.openmuc.jdlms.internal.security.HlsSecretProcessor;

@CosemClass(id = 15, version = 1)
public class AssociationLnClass extends CosemInterfaceObject {

    @CosemAttribute(id = 2, accessMode = AttributeAccessMode.READ_ONLY, type = Type.ARRAY)
    private DataObject objectList;

    @CosemAttribute(id = 3, accessMode = AttributeAccessMode.READ_ONLY, type = Type.STRUCTURE)
    private DataObject associationPartners;

    @CosemAttribute(id = 4, accessMode = AttributeAccessMode.READ_ONLY)
    private DataObject applicationContextName;

    @CosemAttribute(id = 5, accessMode = AttributeAccessMode.READ_ONLY)
    private DataObject authenticationMechanism;

    @CosemAttribute(id = 6, accessMode = AttributeAccessMode.READ_ONLY, type = Type.STRUCTURE)
    private DataObject xDlmsContextInfo;

    @CosemAttribute(id = 7, accessMode = AttributeAccessMode.AUTHENTICATED_READ_AND_WRITE, type = Type.OCTET_STRING)
    private DataObject secret;

    @CosemAttribute(id = 8, accessMode = AttributeAccessMode.READ_ONLY, type = Type.ENUMERATE)
    private DataObject assotiationStatus;

    @CosemAttribute(id = 9, accessMode = AttributeAccessMode.READ_ONLY, type = Type.OCTET_STRING)
    private DataObject securitySetupReference;

    @DlmsDataDirectory
    private DataDirectory dataDirectory;

    private final int logicalDeviceId;

    public AssociationLnClass(int logicalDeviceId) {
        super(WellKnownInstanceIds.ASSOCIATION_LN_ID);
        this.logicalDeviceId = logicalDeviceId;
    }

    public DataObject getObjectList() {
        if (this.objectList != null) {
            return objectList;
        }

        DlmsLogicalDevice logicalDevice = this.dataDirectory.getLogicalDeviceFor(this.logicalDeviceId);
        List<DataObject> objectListElements = new LinkedList<>();

        for (ObisCode instanceId : logicalDevice.getInstanceIds()) {
            CosemClassInstance dlmsClassInstance = logicalDevice.get(instanceId);

            CosemClass dlmsClass = dlmsClassInstance.getCosemClass();

            DataObject classId = DataObject.newUInteger16Data((short) dlmsClass.id());
            DataObject version = DataObject.newUInteger8Data((short) dlmsClass.version());

            List<DataObject> listElement = new LinkedList<>();
            listElement.add(classId);
            listElement.add(version);
            listElement.add(DataObject.newOctetStringData(instanceId.bytes()));

            Collection<Attribute> entries = dlmsClassInstance.getAttribute();
            DataObject attributAccesItems = buildAttributeAccessElementsStruct(entries);

            Collection<MethodAccessor> methods = dlmsClassInstance.getMethods();
            DataObject methodAccesItems = buildMethodAccessElementsStruct(methods);

            List<DataObject> accesRights = new ArrayList<>(2);
            accesRights.add(attributAccesItems);
            accesRights.add(methodAccesItems);

            listElement.add(DataObject.newStructureData(accesRights));

            objectListElements.add(DataObject.newStructureData(listElement));

        }

        this.objectList = DataObject.newArrayData(objectListElements);

        return objectList;
    }

    private DataObject buildMethodAccessElementsStruct(Collection<MethodAccessor> methods) {
        List<MethodAccessor> sortedMethods = sortMethods(methods);

        List<DataObject> attributeAccessItems = new LinkedList<>();
        for (MethodAccessor actionMethod : sortedMethods) {

            CosemMethod methodProperties = actionMethod.getDlmsMethod();

            List<DataObject> attributeAccessItem = new LinkedList<>();
            attributeAccessItem.add(DataObject.newInteger8Data(methodProperties.id()));
            attributeAccessItem.add(DataObject.newEnumerateData((int) methodProperties.accessMode().getCode()));

            attributeAccessItems.add(DataObject.newStructureData(attributeAccessItem));
        }
        return DataObject.newArrayData(attributeAccessItems);
    }

    private DataObject buildAttributeAccessElementsStruct(Collection<Attribute> entries) {
        List<Attribute> sortedEntries = sortAttributes(entries);

        List<DataObject> attributeAccessItems = new LinkedList<>();
        for (Attribute entry : sortedEntries) {

            CosemAttribute attributeProperties = entry.getAttributeProperties();

            List<DataObject> attributeAccessItem = new LinkedList<>();
            attributeAccessItem.add(DataObject.newInteger8Data(attributeProperties.id()));
            attributeAccessItem.add(DataObject.newEnumerateData((int) attributeProperties.accessMode().getCode()));
            attributeAccessItem.add(DataObject.newNullData()); // TODO: check
                                                               // access
                                                               // selectors..

            attributeAccessItems.add(DataObject.newStructureData(attributeAccessItem));
        }

        return DataObject.newArrayData(attributeAccessItems);
    }

    private List<Attribute> sortAttributes(Collection<Attribute> entries) {
        List<Attribute> sortedEntries = new ArrayList<>(entries);
        Collections.sort(sortedEntries, new Comparator<Attribute>() {

            @Override
            public int compare(Attribute o1, Attribute o2) {
                return Integer.compare(o1.getAttributeProperties().id(), o2.getAttributeProperties().id());
            }
        });
        return sortedEntries;
    }

    private List<MethodAccessor> sortMethods(Collection<MethodAccessor> methods) {
        List<MethodAccessor> sortedEntries = new ArrayList<>(methods);
        Collections.sort(sortedEntries, new Comparator<MethodAccessor>() {

            @Override
            public int compare(MethodAccessor o1, MethodAccessor o2) {
                return Integer.compare(o1.getDlmsMethod().id(), o2.getDlmsMethod().id());
            }
        });
        return sortedEntries;
    }

    public DataObject getApplicationContextName() {
        if (applicationContextName == null) {
            byte[] conformance = this.dataDirectory.getLogicalDeviceFor(this.logicalDeviceId)
                    .getLogicalDevice()
                    .getConformance().value;

            this.applicationContextName = DataObject.newOctetStringData(conformance);
        }

        return this.applicationContextName;
    }

    @CosemMethod(id = 1, consumes = Type.OCTET_STRING, accessMode = MethodAccessMode.ACCESS)
    public DataObject replyToHlsAuthentication(DataObject dataObject, Long connectionId)
            throws IllegalMethodAccessException {
        LogicalDevice logicalDevice = this.dataDirectory.getLogicalDeviceFor(this.logicalDeviceId).getLogicalDevice();

        ServerConnectionData connectionData = this.dataDirectory.getConnectionData(connectionId);

        SecuritySuite sec = logicalDevice.getRestrictions().get(connectionData.clientId);

        checkChallengeEquality(dataObject, connectionData);

        HlsSecretProcessor secretProcessor = hlsSecretProcessorFor(sec.getAuthenticationMechanism());

        byte[] clientToServerChallenge = connectionData.clientToServerChallenge;

        byte[] processedClientToServerChallenge;
        try {
            processedClientToServerChallenge = secretProcessor.process(clientToServerChallenge,
                    sec.getAuthenticationKey(), sec.getGlobalUnicastEncryptionKey(), logicalDevice.getSystemTitle(),
                    connectionData.frameCounter);
        } catch (UnsupportedOperationException | IOException e) {
            throw new IllegalMethodAccessException(MethodResultCode.OTHER_REASON);
        }
        connectionData.authenticated = true;

        return DataObject.newOctetStringData(processedClientToServerChallenge);
    }

    private HlsSecretProcessor hlsSecretProcessorFor(AuthenticationMechanism authenticationMechanismId)
            throws IllegalMethodAccessException {

        HlsSecretProcessor secretProcessor;
        switch (authenticationMechanismId) {
        case HLS5_GMAC:
            secretProcessor = new HlsProcessorGmac();
            break;

        default:
        case LOW:
        case NONE:
            // TODO set correct error flag
            throw new IllegalMethodAccessException(MethodResultCode.OTHER_REASON);
        }

        return secretProcessor;
    }

    private void checkChallengeEquality(DataObject dataObject, ServerConnectionData connectionData)
            throws IllegalMethodAccessException {
        byte[] receivedServerToClientChallenge = dataObject.getValue();

        byte[] processedServerToClientChallenge = connectionData.processedServerToClientChallenge;

        if (!Arrays.equals(receivedServerToClientChallenge, processedServerToClientChallenge)) {
            throw new IllegalMethodAccessException(MethodResultCode.OTHER_REASON);
        }
    }
}
