package org.openmuc.jdlms.internal.transportlayer;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;

import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.lnassociation.Association;
import org.openmuc.jdlms.internal.sessionlayer.ServerSessionLayer;
import org.openmuc.jdlms.internal.sessionlayer.ServerWrapperLayer;
import org.openmuc.jdlms.internal.settings.TcpServerSettings;
import org.openmuc.jdlms.transportlayer.client.StreamAccessor;

public class ServerTcpLayer implements ServerTransportLayer {

    private SocketListener socketListener;
    private final DataDirectory directory;
    private ExecutorService singleThreadExecutor;
    private final TcpServerSettings settings;

    public ServerTcpLayer(DataDirectory directory, TcpServerSettings settings) {
        this.directory = directory;
        this.settings = settings;
    }

    @Override
    public void close() throws IOException {
        this.socketListener.close();
        this.singleThreadExecutor.shutdown();
    }

    @Override
    public void start() throws IOException {
        this.socketListener = new SocketListener();
        singleThreadExecutor = Executors.newSingleThreadExecutor();
        singleThreadExecutor.submit(socketListener);
    }

    @Override
    protected void finalize() throws Throwable {
        close();
        super.finalize();
    }

    public static class TcpServerConnectionInformation extends ServerConnectionInformationImpl {

        public InetAddress clienInetAddress;

        public TcpServerConnectionInformation(InetAddress clienInetAddress) {
            this.clienInetAddress = clienInetAddress;
        }

        @Override
        public InetAddress getClienInetAddress() {
            return this.clienInetAddress;
        }
    };

    private class SocketListener implements Runnable, AutoCloseable {

        private final ServerSocket serverSocket;
        private final ThreadPoolExecutor threadPool;
        private final int maxPermits;

        private long connections;

        public SocketListener() throws IOException {
            this.serverSocket = new ServerSocket(settings.tcpPort);
            this.serverSocket.setSoTimeout(0);

            this.maxPermits = settings.maxClients == 0 ? Integer.MAX_VALUE : settings.maxClients;
            this.threadPool = (ThreadPoolExecutor) Executors.newCachedThreadPool();
            this.connections = 0L;
        }

        @Override
        public void run() {
            try {
                while (true) {
                    Socket socket = this.serverSocket.accept();

                    if (this.threadPool.getActiveCount() >= this.maxPermits) {
                        try {
                            socket.close();
                        } catch (IOException e) {
                            // ignore
                        }
                        continue;
                    }

                    ServerSessionLayer sessionLayer = new ServerWrapperLayer(new TcpSocketAccessor(socket), settings);

                    Long connectionId = ++connections;

                    Association association = new Association(directory, sessionLayer, connectionId, settings,
                            new TcpServerConnectionInformation(socket.getInetAddress()));

                    this.threadPool.execute(association);
                }
            } catch (IOException e) {
                // ignore here, connection will be closed
            }

        }

        @Override
        public void close() throws IOException {
            this.serverSocket.close();
            this.threadPool.shutdown();
        }
    }

    private class TcpSocketAccessor implements StreamAccessor {

        private final Socket socket;

        public TcpSocketAccessor(Socket socket) {
            this.socket = socket;
        }

        @Override
        public void setTimeout(int timeout) throws IOException {
            this.socket.setSoTimeout(timeout);
        }

        @Override
        public InputStream getInputStream() throws IOException {
            return this.socket.getInputStream();
        }

        @Override
        public OutputStream getOutpuStream() throws IOException {
            return this.socket.getOutputStream();
        }

        @Override
        public void close() throws IOException {
            this.socket.close();
        }

    }
}
